(* Red-Black trees from
 * http://www.cs.cornell.edu/courses/cs3110/2009sp/lectures/lec11.html
 * see also
 * http://www.lsv.fr/~haddad/transparents-Algo-donnees.pdf
 *
 * Binary search tree with the following properties:
 * A. Nodes are colored either red or black.
 * B. Leaves are black and do not carry elements.
 * C. The root is black
 * D. The children of every red node are black.
 * E. All paths from a node to a descendant leaf have the same
 *    number of black nodes.
 *
 * Longest path from root to leaf can be at most 2x longer than
 * shortest path because at most every other node can be red (no 
 * red node with red parents) and same number of black nodes on
 * each path, thus search always O(log n).
 *)

type 'a tree = [
  | `Leaf
  | `Black of 'a * 'a tree * 'a tree
  | `Red of 'a * 'a tree * 'a tree
]

type 'a rbtree = [
  | `Leaf
  | `Black of 'a * 'a rbtree * 'a rbtree
  | `Red of 'a * 'a btree * 'a btree
] and 'a btree = [
  | `Leaf
  | `Black of 'a * 'a rbtree * 'a rbtree
] and 'a rtree = [
  | `Red of 'a * 'a btree * 'a btree
]

(** Note: on ne peut écrire rbtree = [ rtree | btree ]
  * dans la définition récursive car les types rtree et btree
  * ne sont pas encore définis. *)

(** Exercice:
  * Définir le type des arbres non-vides, et raffiner le typage de
  * l'insertion pour montrer que le résultat est un arbre non-vide. *)

(** [mem] se code plus naturellement sur [tree], cela évite des coercions. *)
let rec mem x : 'a tree -> bool = function
  | `Leaf -> false
  | `Red (y, left, right)
  | `Black (y, left, right) ->
      x = y || (x < y && mem x left) || (x > y && mem x right)

(** Une simple coercion permet de passer de btree (resp. rtree) à rbtree,
  * mais il faut une double coercion pour obtenir un tree. *)
let mem x t = mem x (t : 'a btree :> 'a tree)

(** Pour faire passer l'insertion de façon typée, il faut décortiquer
  * un peu [ins] pour comprendre quel type on va pouvoir lui donner.
  *
  * Quand on insère un nouveau noeud rouge, on brise potentiellement
  * l'invariant rouge. Si on a une chaine noir-rouge-rouge, on rééquilibre,
  * ce qui change le noeud noir en rouge, et peut provoquer une nouvelle
  * violation, etc. On redit cela du point de vue du noeud grand-parent,
  * sous lequel on insère avant de re-équilibrer:
  *
  * - Quand on insère sous un noeud noir, on va toujours pouvoir produire
  *   (après rééquilibrage) un arbre valide, de racine rouge ou noire.
  *
  * - Quand on insère sous un noeud rouge on va produire un arbre dont
  *   le noeud racine est "sale", violant potentiellement l'invariant rouge.
  *   La violation devra être réparée au niveau du noeud père, qui est
  *   nécessairement noir.
  *
  * La fonction [ins] originale réalise (en gros) [btree->rbtree] et
  * [rtree->dirty]. Mais on ne peut l'exprimer directement. Si on approxime
  * en [rbtree->dirty] alors on va se retrouver à rajouter un noeud rouge
  * au dessus de deux autres déja en violation.
  *
  * On s'en sort en découpant l'insertion en deux fonctions,
  * [ins_b : btree->rbtree] et [ins_rb : rbtree -> dirty].
  * La fonction [ins_rb] ne fait jamais d'appel récursif à elle-même,
  * grâce à l'invariant original sur l'arbre, et cela permet d'assurer
  * qu'on n'accumule pas plus de deux noeuds rouges consécutifs. *)

(** Type d'un résultat intermédiaire d'insertion, avec une racine
  * qui viole potentiellement l'invariant rouge. *)
type 'a dirty = [
  | 'a rbtree
  | `RedL of 'a * 'a rtree * 'a btree
  | `RedR of 'a * 'a btree * 'a rtree
]

(** La fonction de ré-équilibrage [balance] est éclatée en quatre:
  *
  *  - Les fonctions black_* font le ré-équilibrage à proprement parler,
  *    selon que le sous-arbre gauche ou droit puisse être dirty.
  *    En les séparant on élimine le cas inutile où les deux sont dirty.
  *
  *  - Les fonctions red_* ne font que rajouter un noeud rouge, mais
  *    de façon à indiquer quand il y a une violation. Elles sont
  *    équivalentes au code précédent dans le cas rouge, modulo les
  *    constructeurs différenciés pour les cas invalides.
  *
  * La séparation de ces cas permet de bien typer, et rend le code
  * plus efficace puisque des cas sont éliminés statiquement. *)

let black_left : 'a * 'a dirty * 'a rbtree -> 'a rbtree = function
  | z, `RedL (y, `Red (x, a, b), c), d
  | z, `RedR (x, a, `Red (y, b, c)), d ->
      let a = (a :> 'a rbtree) in
      let b = (b :> 'a rbtree) in
      let c = (c :> 'a rbtree) in
      `Red (y, `Black (x, a, b), `Black (z, c, d))
  | b, (#rbtree as c), d -> `Black (b,c,d)

let black_right : 'a * 'a rbtree * 'a dirty -> 'a rbtree = function
  | x, a, `RedL (z, `Red (y, b, c), d)
  | x, a, `RedR (y, b, `Red (z, c, d)) ->
      let b = (b :> 'a rbtree) in
      let c = (c :> 'a rbtree) in
      let d = (d :> 'a rbtree) in
        `Red (y, `Black (x, a, b), `Black (z, c, d))
  | b, c, (#rbtree as d) -> `Black (b,c,d)

let red_left : 'a * 'a rbtree * 'a btree -> 'a dirty = function
  | x, (#btree as left), right -> `Red (x,left,right)
  | x, `Red (y,a,b), right -> `RedL (x,`Red (y,a,b),right)

let red_right : 'a * 'a btree * 'a rbtree -> 'a dirty = function
  | x, left, (#btree as right) -> `Red (x,left,right)
  | x, left, `Red (y,a,b) -> `RedR (x,left,`Red (y,a,b))

(** La fonction [insert], avec son [ins] éclaté en [ins_b] et [ins_rb]
  * comme décrit plus haut. *)
let insert x s : 'a btree =
  let rec ins_b : 'a btree -> 'a rbtree = function
    | `Leaf -> `Red (x, `Leaf, `Leaf)
    | `Black (y, a, b) as s ->
        if x < y then black_left (y, ins_rb a, b)
        else if x > y then black_right (y, a, ins_rb b)
        else s
  and ins_rb : 'a rbtree -> 'a dirty = function
    | #btree as t -> (ins_b t :> 'a dirty)
    | `Red (y, a, b) as s ->
        if x < y then red_left (y, ins_b a, b)
        else if x > y then red_right (y, a, ins_b b)
        else s
  in
    match ins_b s with
      | `Red (y,a,b) ->
          let a = (a :> 'a rbtree) in
          let b = (b :> 'a rbtree) in
            `Black (y,a,b)
      | #btree as x -> x

(** Check invariants and test *)

(** Compute number of black nodes along every path from root to leaves.
  * Return this value if it is the same for all paths, [-1] otherwise. *)
let rec path_num_blacks = function
    `Leaf -> 1
  | `Black (_, a, b) | `Red (_, a, b) as t ->
      let ba = path_num_blacks a
      and bb = path_num_blacks b in
        if ba <> bb || ba = -1 || bb = -1
        then -1
        else ba + (match t with `Black _ -> 1 | _ -> 0)

let rec ordered = function
  | `Leaf -> true
  | `Red (v,l,r) | `Black (v,l,r) ->
      ordered l && ordered r &&
      (match l with `Red (w,_,_) | `Black (w,_,_) -> w < v | _ -> true) &&
      (match r with `Red (w,_,_) | `Black (w,_,_) -> v < w | _ -> true)

let is_rbtree (t : 'a btree) =
  (* Condition à la racine et sur les noeuds rouges déja vérifiées.
   * On simplifie le type avant de vérifier les autres invariants. *)
  let t = (t : 'a btree :> 'a tree) in
    ordered t && path_num_blacks t > -1

let rec rbt_of_list : 'a list -> 'a btree = function
  | [] -> `Leaf
  | x :: l -> insert x (rbt_of_list l)

let () =
  let r0 = rbt_of_list [3;5] in
  let r1 = rbt_of_list [3;9;5;7;11] in
  let r2 = rbt_of_list [3;5;7;9;11] in
  let r3 =
    rbt_of_list [1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20]
  in
  let r4 =
    rbt_of_list (Array.to_list (Array.init 84 (fun _ -> Random.int 84)))
  in
  let ra = rbt_of_list ["a";"c";"e";"g";"j"] in
    assert (is_rbtree r0) ;
    assert (is_rbtree r1) ;
    assert (is_rbtree r2) ;
    assert (is_rbtree r3) ;
    assert (is_rbtree r4) ;
    assert (is_rbtree ra) ;
    assert (mem 5 r1 && mem 9 r1 && not (mem 6 r1)) ;
    Printf.printf "All tests passed successfully.\n"
